/*
 *  typespeed - measures your typing speed
 *  Copyright (C) 1999-2003   Jani Ollikainen  <bestis@iki.fi>
 *                          & Jaakko Manelius  <jman@iki.fi>
 *  Copyright (C) 2006-2007   Tobias Stoeckmann  <tobias@bugol.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * typespeed.c - the main code
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/types.h>

#include <ctype.h>
#include <curses.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "pathnames.h"
#include "typespeed.h"
#include "version.h"

#define DEFAULT_CHEAT		0
#define DEFAULT_CLIENT		0
#define DEFAULT_COLORS		1
#define DEFAULT_PORT		6025
#define DEFAULT_SERVER		0

extern int		addtop10(char *, struct stats *);
extern int		choosewordfile(int);
extern void		closenetwork(void);
extern void		drawmenu(void);
extern void		drawscreen(void);
extern void		drawstatus(unsigned);
extern void		endcursestuff(void);
extern void		freewords(void);
extern void		initcursestuff(void);
extern int		initnetwork(char *, int);
extern unsigned short	level(int);
extern void		loadhighscores(char *);
extern void		makescorefiles(char *);
extern void		multipmenu(void);
extern int		netrecvword(int);
extern void		netsendword(int, char *);
extern void		netswapscore(int, struct stats *, struct stats *);
extern void		netwaitgame(int);
extern void		optionmenu(void);
extern void		pressanykey(int, int);
extern int		r(int);
extern void		readconfig(void);
extern void		rulesmenu(void);
extern void		showhighscores(void);
extern void		tellstory(void);
extern clock_t		timenow(void);
extern int		typorankkaus(float);
extern void		xcolor_set(short);
extern void		xerr(int, const char *, ...);
extern void		xerrx(int, const char *, ...);

int			addnewword(char *);
static void		clearword(int, int, size_t);
static void		cwords(void);
void			defrule(void);
static void		movewords(void);
static inline void	parseinput(int *, char *, clock_t *, unsigned *, int *);
static int		play(void);
static void		usage(void);

/* globals */
char *rankki[11] =
{"None", "Loser", "ADP-Pro", "NoGood", "Average",
 "Good", "VeryGood", "Pro", "3l33t", "*(GOD)*", "Computer"};

char *typorank[12] =
{"None", "Alien", "Secretary", "Human", "IT Person", "Handicap",
 "Monkey", "Pencil", "T-Bone", "E-Typo", "TypOmatiC", "TypoKing"};

struct stats	 best;
int		 hfd;
int		 misses;
struct stats	 other;
char		*progname;
float		 rate;
int		 typesocket = 0;
char		*usedwordfile;
char		*word[WORD_MAX];
int		 wordcount = 0;
int		 wordpos[22];
char		 wordstring[22][20];

struct option options[] = {
	{"cheat", no_argument, &opt.cheat, 1},
	{"client", required_argument, NULL, 'o'},
	{"help", no_argument, NULL, 'h'},
	{"nocolors", no_argument, &opt.usecolors, 0},
	{"port", required_argument, NULL, 'p'},
	{"server", no_argument, &opt.server, 1},
	{0, 0, 0, 0}
};

/*
 * Insert a new word into wordstring and wordpos, i.e. one more word
 * on the screen.
 *
 * If newword points to a string, newword will be inserted.
 * If newword is NULL, a random word out of word will be inserted.
 * In both cases, duplicates are not allowed. Should newword point
 * to duplicated string or string that cannot be typed with current
 * locale, no string will be inserted.
 *
 * Returns TRUE on success or FALSE on failure.
 *
 * XXX: Can result in a veeery long loop when there are many words on
 *      field and only a few words in word list (no duplicates allowed).
 */
int
addnewword(char *newword)
{
	int count, dup, slot;
	int freeslot[22];
	size_t i;
	char *myword, *p;

	if (newword != NULL) {
		if (strlen(newword) > (size_t)rules.maxlen)
			return FALSE;
		for (p = newword; *p != '\0'; p++)
			if (!isprint(*p) || isspace(*p))
				return FALSE;
	}

	/* take a random free slot */
	for (count = 0, i = 0; i < 22; i++)
		if (wordpos[i] == -2)
			freeslot[count++] = i;

	if (!count || count < 23 - rules.maxwords)
		/* the player has enough trouble... */
		return FALSE;

	slot = freeslot[r(count)];

	if ((myword = newword) == NULL)
		myword = word[r(wordcount)];

	do {
		for (dup = 0, i = 0; i < 22; i++)
			if (!strcmp(wordstring[i], myword)) {
				if (newword != NULL)
					return FALSE;
				dup = 1;
				myword = word[r(wordcount)];
				break;
			}
	} while (dup);

	(void)strncpy(wordstring[slot], myword, sizeof(wordstring[slot]) - 1);
	wordpos[slot] = -1;

	return TRUE;
}

/* Writes "length" blanks at position x y. */
static void
clearword(int y, int x, size_t length)
{
	move(y, x);
	while (length--)
		addch(' ');
}

/*
 * Sets rate according to current length of all visible words.
 * If words with too few chars are on screen, throw in another word.
 */
static void
cwords(void)
{
	int i, wc;
	unsigned long length;

	for (wc = 0, length = 0, i = 0; i < 22; i++)
		if (wordpos[i] > -2) {
			wc++;
			length += strlen(wordstring[i]) + 1;
		}

	if (length < now.score / 4 + 1 || wc < rules.minwords)
		addnewword(NULL);

	if (rules.smooth)
		rate = (float)now.score / rules.step + rules.minspeed;
	else
		rate = now.score / rules.step + rules.minspeed;

	if (rules.maxspeed && rate > rules.maxspeed)
		rate = rules.maxspeed;
}

/*
 * Resets rules to typespeed's default.
 */
void
defrule(void)
{
	rules.misses = 10;
	rules.minlen = 1;
	rules.maxlen = 19;
	rules.minwords = 1;
	rules.maxwords = 22;
	rules.hightype = 1;
	rules.minscore = 0;
	rules.minspeed = 3;
	rules.maxspeed = 0;
	rules.step = 175;
	rules.smooth = 1;
	(void)strncpy(rules.fname, "default", sizeof(rules.fname) - 1);
	rules.fname[sizeof(rules.fname) - 1] = '\0';
	(void)strncpy(rules.name, "default", sizeof(rules.name) - 1);
	rules.name[sizeof(rules.name) - 1] = '\0';
}

#ifndef TEST
int
main(int argc, char **argv)
{
	int i, wheretogo;
	gid_t gid;
	char serv[MAXHOSTNAMELEN];

	if ((progname = strrchr(argv[0], '/')) == NULL)
		progname = argv[0];
	else
		progname++;

	/* just open high score file while being setgid games */
	if ((hfd = open(HIGHFILE, O_RDWR, 0)) == -1)
		xerr(1, "main: open: %s", HIGHFILE);

	/* drop privileges */
	gid = getgid();
	setresgid(gid, gid, gid);

	/* check file descriptors for consistency */
	if (hfd == STDIN_FILENO || hfd == STDOUT_FILENO ||
	    hfd == STDERR_FILENO)
		exit(1);
	if (!isatty(STDIN_FILENO) || !isatty(STDOUT_FILENO) ||
	    !isatty(STDERR_FILENO))
		xerrx(1, "not fully connected to a terminal");

	if (setlocale(LC_CTYPE, "") == NULL)
		xerrx(1, "Cannot set locale");

	opt.cheat = DEFAULT_CHEAT;
	opt.client = DEFAULT_CLIENT;
	opt.port = DEFAULT_PORT;
	opt.server = DEFAULT_SERVER;
	opt.usecolors = DEFAULT_COLORS;

	usedwordfile = NULL;

	defrule();

	memset(&best, 0, sizeof(best));
	memset(&now, 0, sizeof(now));
	memset(&word, 0, sizeof(word));

	readconfig();

	printf("Typespeed %s, Copyright (C) "
	    "Jani Ollikainen & Jaakko Manelius\n"
	    "                  Copyright (C) Tobias Stoeckmann\n"
	    "Typespeed comes with ABSOLUTELY NO WARRANTY;"
	    "for details read COPYING.\n",
	    TVERSION);

	while ((i = getopt_long(argc, argv, "to:h", options, NULL)) != -1) {
		switch (i) {
		case 0:
			/* flag already set by getopt */
			break;
		case 'o':
			opt.client = 1;
			if (!optarg)
				/* getopt should handle this */
				abort();
			(void)strncpy(serv, optarg, sizeof(serv) - 1);
			serv[sizeof(serv) - 1] = '\0';
			break;
		case 'p':
			if (!optarg)
				/* getopt should handle this */
				abort();
			if ((opt.port = atoi(optarg)) < 1025) {
				fprintf(stderr, "Do not use privileged port!");
				exit(1);
			}
			break;
		case ':':
		case '?':
			/* getopt already printed error */
		case 'h':
		default:
			usage();
			/* NOTREACHED */
			break;
		}
	}

	if (opt.server || opt.client)
		while ((typesocket = initnetwork(serv, 0)) < 0)
			;

	sleep(1);

	initcursestuff();

	if (typesocket > 0)
		play();

	drawmenu();
	/*
	 * THIS IS OLD NEWS - LEFT HERE TO PUT SOME FUTURE NEWS IN
	 * mvaddstr(15, 20, "...");
	 * mvaddstr(16, 26, "...");
	 */

	wheretogo = 0;
	do {
		nodelay(stdscr, 0);
		cbreak();
	
		while ((wheretogo = getch()) == ERR)
			;

		switch (wheretogo) {
		case '1':
			play();
			break;
		case '2':
			multipmenu();
			if (typesocket > 0)
				play();
			break;
		case '3':
			tellstory();
			break;
		case '4':
			showhighscores();
			break;
		case '5':
			optionmenu();
			break;
		case '6':
			rulesmenu();
			break;
		default:
			break;
		}
		if (wheretogo > '0' && wheretogo < '7')
			drawmenu();
	} while (wheretogo != '7');

	endcursestuff();

	printf("\033[H\033[J"
	    "Typespeed %s\n\n"
	    "Best score was:\n\n"
	    "Rank:\t\t%s\n"
	    "Score:\t\t%lu\n"
	    "Total CPS:\t%2.3f\n"
	    "Correct CPS:\t%2.3f\n"
	    "Typo ratio:\t%2.1f%%\n"
	    "Typorank:\t",
	    TVERSION, rankki[best.level], best.score, best.totalspeed,
	    best.speed, best.ratio);
	if (best.tcount)
		puts(typorank[typorankkaus(best.ratio)]);
	else
		puts("None");
	fputc('\n', stdout);
	return 0;
}
#endif /* TEST */

/*
 * Move words across the screen.
 * Removes them if they reach the right side and increases misses.
 */
static void
movewords(void)
{
	int i, length;

	for (i = 0; i < 22; i++) {
		if (wordpos[i] == -2)
			continue;

		length = (int)strlen(wordstring[i]);
		xcolor_set(0);

		if (wordpos[i] > (79 - length)) {
			clearword(i, wordpos[i], length);
			misses++;
			wordpos[i] = -2;
			continue;
		}

		if (wordpos[i] > -1)
			mvaddch(i, wordpos[i], ' ');

		if (wordpos[i] > (65 - length))
			xcolor_set(3);
		else if (wordpos[i] > (50 - length))
			xcolor_set(7);
		else
			xcolor_set(1);

		mvaddstr(i, ++wordpos[i], wordstring[i]);
	}
}

/*
 * Parses user input during game.
 */
static inline void
parseinput(int *escend, char *input, clock_t *starttime, unsigned *inputpos,
    int *nappaykset)
{
	int i, nappi;
	clock_t pausetime;

	if ((nappi = getch()) == ERR)
		return;

	switch (tolower(nappi)) {
	case 8:
	case 127:
	case KEY_BACKSPACE:
		if (*inputpos) {
			mvaddch(23, 1 + *inputpos, ' ');
			input[--(*inputpos)] = '\0';
		}
		break;
	case 27: /* ESC */
		*escend = 1;
		break;
	case KEY_UP:
		if (opt.client || opt.server) {
			*escend = 1;
			break;
		}

		pausetime = timenow();
		mvaddstr(23, 2, "       PAUSED      ");
		move(23, 2);
		while (getch() == ERR)
			;
		*starttime += timenow() - pausetime;
		mvaddstr(23, 9, "      ");
		flushinp();
		/* FALLTHROUGH */
	case 21: /* ^U */
		if (*inputpos) {
			clearword(23, 2, *inputpos);
			*inputpos = 0;
			input[0] = '\0';
		}
		break;
	case 32: /* SPACE */
	case 10: /* ENTER */
		for (i = 0; i < 22; i++) {
			if (strcmp(input, wordstring[i]))
				continue;
			now.wordswritten++;
			now.score += *inputpos;
			clearword(i, wordpos[i], strlen(wordstring[i]));
			wordpos[i] = -2;
			/* send ~ 50% words in multiplayer */
			if ((opt.server || opt.client) && !r(2))
				netsendword(typesocket, wordstring[i]);
			strncpy(wordstring[i], " ", sizeof(wordstring[i]) - 1);
		}
		clearword(23, 2, 19);
		*inputpos = 0;
		input[0] = '\0';
		break;
	case 11: /* ^K */
		if ((size_t)*inputpos < strlen(input)) {
			input[*inputpos] = '\0';
			clearword(23, 2 + *inputpos, 19 - *inputpos);
		}
		break;
	case KEY_RIGHT:
	case 6: /* ^F */
		if (*inputpos < 19 && (size_t)*inputpos < strlen(input))
			(*inputpos)++;
		break;
	case KEY_END:
	case 5: /* ^E */
		*inputpos = strlen(input);
		break;
	case KEY_LEFT:
	case 2: /* ^B */
		if (*inputpos)
			(*inputpos)--;
		break;
	case KEY_HOME:
	case 1: /* ^A */
		*inputpos = 0;
		break;
	default:
		if (nappi > 255)
			break;
		(*nappaykset)++;
		if (*inputpos > 18 || iscntrl(nappi))
			break;
		mvaddch(23, 2 + *inputpos, nappi);
		if (*inputpos == strlen(input))
			input[*inputpos + 1] = '\0';
		input[(*inputpos)++] = nappi;
		if (!opt.cheat)
			break;
		for (i = 0; i < 22; i++) {
			if (strcmp(input, wordstring[i]))
				continue;
			now.score += *inputpos;
			clearword(i, wordpos[i], strlen(wordstring[i]));
			wordpos[i] = -2;
			clearword(23, 2, 19);
			*inputpos = 0;
			input[0]= '\0';
			strncpy(wordstring[i], " ", sizeof(wordstring[i]) - 1);
		}
		break;
	}
	move(23, 2 + *inputpos);
}

/*
 * Contains core game logic.
 * This functions runs as long player hasn't reached main screen again. :)
 */
static int
play(void)
{
	int escend, i, nappaykset;
	unsigned inputpos;
	float min10;
	clock_t oldtimes, starttime;
	char input[20];

	now.sinit = time(NULL);
	srandom((unsigned long)now.sinit);

	if (opt.client || opt.server)
		defrule();

	while (choosewordfile(0))
		;

	misses = 0;
	min10 = 0;
	escend = inputpos = nappaykset = 0;
	rate = 1;

	memset(&now, 0, sizeof(now));

	drawscreen();
	nodelay(stdscr, 1);
	cbreak();

	if (opt.server || opt.client) {
		xcolor_set(2);
		mvaddstr(12, 20, "Waiting for the other party to join in...");
		refresh();
		netwaitgame(typesocket);
		flushinp();
		drawscreen();
	}

	halfdelay(1);

	oldtimes = starttime = timenow();
	input[0] = '\0';

	for (i = 0; i < 22; i++) {
		wordstring[i][0] = ' ';
		wordstring[i][1] = '\0';
		wordpos[i] = -2;
	}

	/* move cursor to correct position */
	drawstatus(0);

	while (!escend && misses < rules.misses) {
		while (timenow() < (oldtimes + (100.0 / rate)))
			parseinput(&escend, input, &starttime, &inputpos,
			    &nappaykset);
		oldtimes = timenow();

		movewords();
		cwords();

		now.duration = timenow() - starttime;
		now.speed = (now.score + now.wordswritten) * 100.0 /
		    now.duration;
		now.totalspeed = (nappaykset + now.wordswritten) * 100.0 /
		    now.duration;

		drawstatus(inputpos);

		/* last, we check if the opponent has new words for us... */
		if ((opt.server || opt.client) && !escend)
			escend = netrecvword(typesocket);

		refresh();
	}

	if ((opt.client || opt.server) && escend != 2)
		netsendword(typesocket, " ENDOFTHEGAME\n");

	endcursestuff();
	initcursestuff();
	drawscreen();

	now.duration = oldtimes - starttime;
	now.tcount = nappaykset;
	now.level = level(now.score);

	if (now.totalspeed)
		now.ratio = (1 - now.speed / now.totalspeed) * 100;

	if (now.score >= best.score)
		memcpy((void *)&best, (void *)&now, sizeof(best));

	if (opt.server || opt.client)
		netswapscore(typesocket, &now, &other);

	xcolor_set(5);

	mvaddstr(12, 30, "GAME OVER!");
	refresh();

	sleep(3);

	/* total speed */
	min10 = now.speed * 600;

	clearword(12, 30, strlen("GAME OVER!"));

	mvprintw(3, 20, "Typespeed %s", TVERSION);

	mvaddstr(5, 20, "Your score was:");

	mvaddstr(7, 20, "Rank:");
	mvaddstr(7, 40, rankki[level(now.score)]);
	xcolor_set(5);
	mvprintw(8, 20, "Score:\t\t%d", now.score);
	mvprintw(9, 20, "10MRS:\t\t%1.0f", min10);
	mvprintw(10, 20, "Total CPS:\t\t%2.3f", now.totalspeed);
	mvprintw(11, 20, "Correct CPS:\t%2.3f", now.speed);
	mvprintw(12, 20, "Typo ratio:\t\t%2.1f%%", now.ratio);
	mvaddstr(13, 20, "Typorank:");

	if (!now.tcount)
		mvaddstr(13, 40, "None");
	else
		mvaddstr(13, 40, typorank[typorankkaus(now.ratio)]);

	if (opt.client || opt.server) {
		mvaddstr(7, 60, rankki[level(other.score)]);
		xcolor_set(5);
		mvaddstr(5, 60, "Opponent's:");
		mvprintw(8, 60, "%d", other.score);
		mvprintw(9, 60, "%1.0f", other.speed * 600);
		mvprintw(10, 60, "%2.3f", other.totalspeed);
		mvprintw(11, 60, "%2.3f", other.speed);
		if (other.totalspeed)
			other.ratio = (1 - other.speed / other.totalspeed)
			    * 100;
		mvprintw(12, 60, "%2.1f%%", other.ratio);
		if (other.ratio)
			mvaddstr(13, 60, typorank[typorankkaus(other.ratio)]);
		else
			mvaddstr(13, 60, "None");
	}

	xcolor_set(2);
	if (opt.client || opt.server)
		mvaddstr(17, 15, "NETWORK MODE - YOU CANNOT GET TO TOP10");
	else if (!rules.hightype)
		mvaddstr(17, 15, "RANKING DISABLED - YOU CANNOT GET TO TOP10");
	nodelay(stdscr, 0);
	cbreak();
	flushinp();

	pressanykey(15, 20);
	if (!opt.client && !opt.server &&
	    now.score > rules.minscore && rules.hightype) {
		addtop10(usedwordfile, &now);
		loadhighscores(usedwordfile);
		pressanykey(16, 3);
	}
	free(usedwordfile);
	usedwordfile = NULL;
	freewords();
	return 0;
}

/* Prints usage information on stderr and exits with return value 1. */
static void
usage()
{
	fprintf(stderr,
	    "\n"
	    "command line options:\n"
	    "  --cheat       = multimedia cheat(?)\n"
	    "  --client=addr = ip address of server\n"
	    "  --nocolors    = do not use colors\n"
	    "  --port=port   = port number in the network play\n"
	    "  --server      = start typespeed in server mode\n"
	    "  --help        = this help!\n"
	    "\n"
	    "usage: %s options\n"
	    "\n"
	    "typespeed %s - http://tobias.eyedacor.org/typespeed/\n",
	    progname, TVERSION);
	exit(1);
}

